@@ -1,7 +1,5 @@
 import os
 import sys
-from abc import ABC, abstractmethod
-from collections import namedtuple
 from contextlib import contextmanager
 
 import click
@@ -9,23 +7,22 @@ from click import UsageError
 from dagster import check
 from dagster.core.code_pointer import CodePointer
 from dagster.core.definitions.reconstructable import repository_def_from_target_def
-from dagster.core.host_representation import (
-    ExternalRepository,
-    ExternalRepositoryOrigin,
-    GrpcServerRepositoryLocationOrigin,
-    RepositoryLocation,
-)
+from dagster.core.host_representation.external import ExternalRepository
+from dagster.core.host_representation.repository_location import RepositoryLocation
+from dagster.core.instance import DagsterInstance
 from dagster.core.origin import PipelinePythonOrigin, RepositoryPythonOrigin
+from dagster.core.workspace.context import WorkspaceRequestContext
+from dagster.core.workspace.load_target import (
+    EmptyWorkspaceTarget,
+    GrpcServerTarget,
+    ModuleTarget,
+    PackageTarget,
+    PythonFileTarget,
+    WorkspaceFileTarget,
+)
 from dagster.grpc.utils import get_loadable_targets
 from dagster.utils.hosted_user_process import recon_repository_from_origin
 
-from .load import (
-    location_origin_from_module_name,
-    location_origin_from_package_name,
-    location_origin_from_python_file,
-    location_origins_from_yaml_paths,
-)
-
 WORKSPACE_TARGET_WARNING = "Can only use ONE of --workspace/-w, --python-file/-f, --module-name/-m, --grpc-port, --grpc-socket."
 
 
@@ -66,78 +63,6 @@ WORKSPACE_CLI_ARGS = (
 )
 
 
-class WorkspaceLoadTarget(ABC):
-    @abstractmethod
-    def create_origins(self):
-        """ Reloads the RepositoryLocationOrigins for this workspace."""
-
-
-class WorkspaceFileTarget(namedtuple("WorkspaceFileTarget", "paths"), WorkspaceLoadTarget):
-    def create_origins(self):
-        return location_origins_from_yaml_paths(self.paths)
-
-
-class PythonFileTarget(
-    namedtuple("PythonFileTarget", "python_file attribute working_directory location_name"),
-    WorkspaceLoadTarget,
-):
-    def create_origins(self):
-        return [
-            location_origin_from_python_file(
-                python_file=self.python_file,
-                attribute=self.attribute,
-                working_directory=self.working_directory,
-                location_name=self.location_name,
-            )
-        ]
-
-
-class ModuleTarget(
-    namedtuple("ModuleTarget", "module_name attribute location_name"), WorkspaceLoadTarget
-):
-    def create_origins(self):
-        return [
-            location_origin_from_module_name(
-                self.module_name,
-                self.attribute,
-                location_name=self.location_name,
-            )
-        ]
-
-
-class PackageTarget(
-    namedtuple("PackageTarget", "package_name attribute location_name"), WorkspaceLoadTarget
-):
-    def create_origins(self):
-        return [
-            location_origin_from_package_name(
-                self.package_name,
-                self.attribute,
-                location_name=self.location_name,
-            )
-        ]
-
-
-class GrpcServerTarget(
-    namedtuple("GrpcServerTarget", "host port socket location_name"), WorkspaceLoadTarget
-):
-    def create_origins(self):
-        return [
-            GrpcServerRepositoryLocationOrigin(
-                port=self.port,
-                socket=self.socket,
-                host=self.host,
-                location_name=self.location_name,
-            )
-        ]
-
-
-#  Utility target for graphql commands that do not require a workspace, e.g. downloading schema
-class EmptyWorkspaceTarget(namedtuple("EmptyWorkspaceTarget", ""), WorkspaceLoadTarget):
-    def create_origins(self):
-        return []
-
-
 def created_workspace_load_target(kwargs):
     check.dict_param(kwargs, "kwargs")
     if are_all_keys_empty(kwargs, WORKSPACE_CLI_ARGS):
@@ -237,10 +162,22 @@ def created_workspace_load_target(kwargs):
         _cli_load_invariant(False)
 
 
-def get_workspace_from_kwargs(kwargs):
-    from .workspace import Workspace
+def get_workspace_process_context_from_kwargs(
+    instance: DagsterInstance, version: str, read_only: bool, kwargs
+):
+    from dagster.core.workspace import WorkspaceProcessContext
 
-    return Workspace(created_workspace_load_target(kwargs))
+    return WorkspaceProcessContext(
+        instance, created_workspace_load_target(kwargs), version=version, read_only=read_only
+    )
+
+
+@contextmanager
+def get_workspace_from_kwargs(instance: DagsterInstance, version: str, kwargs):
+    with get_workspace_process_context_from_kwargs(
+        instance, version, read_only=False, kwargs=kwargs
+    ) as workspace_process_context:
+        yield workspace_process_context.create_request_context()
 
 
 def python_target_click_options():
@@ -362,7 +299,7 @@ def target_with_config_option(command_name):
             "\n\nYou can also specify multiple files:"
             "\n\nExample: "
             "dagster pipeline {name} -f hello_world.py -p pandas_hello_world "
-            "-c pandas_hello_world/solids.yaml -e pandas_hello_world/env.yaml"
+            "-c pandas_hello_world/solids.yaml -c pandas_hello_world/env.yaml"
         ).format(name=command_name),
     )
 
@@ -436,17 +373,6 @@ def pipeline_target_argument(f):
     return apply_click_params(repository_target_argument(f), pipeline_option())
 
 
-def get_repository_origin_from_kwargs(kwargs):
-    provided_repo_name = kwargs.get("repository")
-
-    if not provided_repo_name:
-        raise click.UsageError("Must provide --repository to load a repository")
-
-    repository_location_origin = get_repository_location_origin_from_kwargs(kwargs)
-
-    return ExternalRepositoryOrigin(repository_location_origin, provided_repo_name)
-
-
 def get_pipeline_python_origin_from_kwargs(kwargs):
     repository_origin = get_repository_python_origin_from_kwargs(kwargs)
     provided_pipeline_name = kwargs.get("pipeline")
@@ -588,43 +514,49 @@ def get_repository_python_origin_from_kwargs(kwargs):
 
 
 @contextmanager
-def get_repository_location_from_kwargs(kwargs):
-    origin = get_repository_location_origin_from_kwargs(kwargs)
-    with origin.create_handle() as handle:
-        yield handle.create_location()
+def get_repository_location_from_kwargs(instance, version, kwargs):
+    # Instance isn't strictly required to load a repository location, but is included
+    # to satisfy the WorkspaceProcessContext / WorkspaceRequestContext requirements
+    with get_workspace_from_kwargs(instance, version, kwargs) as workspace:
+        yield get_repository_location_from_workspace(workspace, kwargs.get("location"))
 
 
-def get_repository_location_origin_from_kwargs(kwargs):
-    all_origins = created_workspace_load_target(kwargs).create_origins()
+def get_repository_location_from_workspace(
+    workspace: WorkspaceRequestContext, provided_location_name
+):
+    if provided_location_name is None:
+        if len(workspace.repository_location_names) == 1:
+            provided_location_name = workspace.repository_location_names[0]
+        elif len(workspace.repository_location_names) == 0:
+            raise click.UsageError("No locations found in workspace")
+        elif provided_location_name is None:
+            raise click.UsageError(
+                (
+                    "Must provide --location as there are multiple locations "
+                    "available. Options are: {}"
+                ).format(_sorted_quoted(workspace.repository_location_names))
+            )
 
-    origins_by_name = {origin.location_name: origin for origin in all_origins}
-
-    provided_location_name = kwargs.get("location")
-
-    if provided_location_name is None and len(origins_by_name) == 1:
-        return next(iter(origins_by_name.values()))
-
-    elif provided_location_name is None:
-        raise click.UsageError(
-            (
-                "Must provide --location as there are more than one locations "
-                "available. Options are: {}"
-            ).format(_sorted_quoted(origins_by_name.keys()))
-        )
-
-    elif not origins_by_name.get(provided_location_name):
+    if provided_location_name not in workspace.repository_location_names:
         raise click.UsageError(
             (
                 'Location "{provided_location_name}" not found in workspace. '
                 "Found {found_names} instead."
             ).format(
                 provided_location_name=provided_location_name,
-                found_names=_sorted_quoted(origins_by_name.keys()),
+                found_names=_sorted_quoted(workspace.repository_location_names),
             )
         )
 
-    else:
-        return origins_by_name.get(provided_location_name)
+    if workspace.has_repository_location_error(provided_location_name):
+        raise click.UsageError(
+            'Error loading location "{provided_location_name}": {error_str}'.format(
+                provided_location_name=provided_location_name,
+                error_str=str(workspace.get_repository_location_error(provided_location_name)),
+            )
+        )
+
+    return workspace.get_repository_location(provided_location_name)
 
 
 def get_external_repository_from_repo_location(repo_location, provided_repo_name):
@@ -662,8 +594,10 @@ def get_external_repository_from_repo_location(repo_location, provided_repo_name
 
 
 @contextmanager
-def get_external_repository_from_kwargs(kwargs):
-    with get_repository_location_from_kwargs(kwargs) as repo_location:
+def get_external_repository_from_kwargs(instance, version, kwargs):
+    # Instance isn't strictly required to load an ExternalRepository, but is included
+    # to satisfy the WorkspaceProcessContext / WorkspaceRequestContext requirements
+    with get_repository_location_from_kwargs(instance, version, kwargs) as repo_location:
         provided_repo_name = kwargs.get("repository")
         yield get_external_repository_from_repo_location(repo_location, provided_repo_name)
 
@@ -705,8 +639,10 @@ def get_external_pipeline_from_external_repo(external_repo, provided_pipeline_na
 
 
 @contextmanager
-def get_external_pipeline_from_kwargs(kwargs):
-    with get_external_repository_from_kwargs(kwargs) as external_repo:
+def get_external_pipeline_from_kwargs(instance, version, kwargs):
+    # Instance isn't strictly required to load an ExternalPipeline, but is included
+    # to satisfy the WorkspaceProcessContext / WorkspaceRequestContext requirements
+    with get_external_repository_from_kwargs(instance, version, kwargs) as external_repo:
         provided_pipeline_name = kwargs.get("pipeline")
         yield get_external_pipeline_from_external_repo(external_repo, provided_pipeline_name)
 
