@@ -3,13 +3,12 @@ from collections import defaultdict, namedtuple
 from enum import Enum
 from typing import (
     TYPE_CHECKING,
+    AbstractSet,
     Any,
     Dict,
-    FrozenSet,
     List,
     NamedTuple,
     Optional,
-    Set,
     Tuple,
     Type,
     Union,
@@ -17,6 +16,7 @@ from typing import (
 )
 
 from dagster import check
+from dagster.core.definitions.policy import RetryPolicy
 from dagster.core.errors import DagsterInvalidDefinitionError
 from dagster.serdes import whitelist_for_serdes
 from dagster.utils import frozentags
@@ -32,7 +32,7 @@ if TYPE_CHECKING:
     from .solid import NodeDefinition
 
 
-class SolidInvocation(namedtuple("Solid", "name alias tags hook_defs")):
+class SolidInvocation(namedtuple("Solid", "name alias tags hook_defs retry_policy")):
     """Identifies an instance of a solid in a pipeline dependency structure.
 
     Args:
@@ -41,7 +41,7 @@ class SolidInvocation(namedtuple("Solid", "name alias tags hook_defs")):
             multiple instances of the same solid.
         tags (Optional[Dict[str, Any]]): Optional tags values to extend or override those
             set on the solid definition.
-        hook_defs (Optional[Set[HookDefinition]]): A set of hook definitions applied to the
+        hook_defs (Optional[AbstractSet[HookDefinition]]): A set of hook definitions applied to the
             solid instance.
 
     Examples:
@@ -78,13 +78,19 @@ class SolidInvocation(namedtuple("Solid", "name alias tags hook_defs")):
         name: str,
         alias: Optional[str] = None,
         tags: Dict[str, str] = None,
-        hook_defs: FrozenSet[HookDefinition] = None,
+        hook_defs: AbstractSet[HookDefinition] = None,
+        retry_policy: Optional[RetryPolicy] = None,
     ):
-        name = check.str_param(name, "name")
-        alias = check.opt_str_param(alias, "alias")
-        tags = frozentags(check.opt_dict_param(tags, "tags", value_type=str, key_type=str))
-        hook_defs = frozenset(check.opt_set_param(hook_defs, "hook_defs", of_type=HookDefinition))
-        return super(cls, SolidInvocation).__new__(cls, name, alias, tags, hook_defs)
+        return super().__new__(
+            cls,
+            name=check.str_param(name, "name"),
+            alias=check.opt_str_param(alias, "alias"),
+            tags=frozentags(check.opt_dict_param(tags, "tags", value_type=str, key_type=str)),
+            hook_defs=frozenset(
+                check.opt_set_param(hook_defs, "hook_defs", of_type=HookDefinition)
+            ),
+            retry_policy=check.opt_inst_param(retry_policy, "retry_policy", RetryPolicy),
+        )
 
 
 class Solid:
@@ -104,7 +110,8 @@ class Solid:
         definition: "NodeDefinition",
         graph_definition: "GraphDefinition",
         tags: Dict[str, str] = None,
-        hook_defs: Optional[Set[HookDefinition]] = None,
+        hook_defs: Optional[AbstractSet[HookDefinition]] = None,
+        retry_policy: Optional[RetryPolicy] = None,
     ):
         from .graph import GraphDefinition
         from .solid import NodeDefinition
@@ -118,6 +125,7 @@ class Solid:
         )
         self._additional_tags = validate_tags(tags)
         self._hook_defs = check.opt_set_param(hook_defs, "hook_defs", of_type=HookDefinition)
+        self._retry_policy = check.opt_inst_param(retry_policy, "retry_policy", RetryPolicy)
 
         input_handles = {}
         for name, input_def in self.definition.input_dict.items():
@@ -198,9 +206,13 @@ class Solid:
         )
 
     @property
-    def hook_defs(self) -> Optional[Set[HookDefinition]]:
+    def hook_defs(self) -> AbstractSet[HookDefinition]:
         return self._hook_defs
 
+    @property
+    def retry_policy(self) -> Optional[RetryPolicy]:
+        return self._retry_policy
+
 
 @whitelist_for_serdes
 class SolidHandle(namedtuple("_SolidHandle", "name parent")):
@@ -536,7 +548,10 @@ class MultiDependencyDefinition(
         description (Optional[str]): Human-readable description of this dependency.
     """
 
-    def __new__(cls, dependencies: List[Union[DependencyDefinition, "MappedInputPlaceholder"]]):
+    def __new__(
+        cls,
+        dependencies: List[Union[DependencyDefinition, Type["MappedInputPlaceholder"]]],
+    ):
         from .composition import MappedInputPlaceholder
 
         deps = check.list_param(dependencies, "dependencies")
@@ -655,8 +670,8 @@ class DependencyStructure:
         # solid_name => dynamic output_handle that this will solid dupe for
         self._dynamic_fan_out_index: dict = {}
 
-        # solid_name => dynamic output_handle this collects over
-        self._collect_index: dict = {}
+        # solid_name => set of dynamic output_handle this collects over
+        self._collect_index: Dict[str, set] = defaultdict(set)
 
         for input_handle, (dep_type, output_handle_or_list) in self._handle_dict.items():
             if dep_type == DependencyType.FAN_IN:
@@ -733,7 +748,7 @@ class DependencyStructure:
             raise DagsterInvalidDefinitionError(
                 f'Solid "{input_handle.solid_name}" cannot be both downstream of dynamic output '
                 f"{output_handle.describe()} and collect over dynamic output "
-                f"{self._collect_index[input_handle.solid_name].describe()}."
+                f"{list(self._collect_index[input_handle.solid_name])[0].describe()}."
             )
 
         if self._dynamic_fan_out_index.get(input_handle.solid_name) is None:
@@ -759,10 +774,7 @@ class DependencyStructure:
                 f"{self._dynamic_fan_out_index[input_handle.solid_name].describe()}."
             )
 
-        if self._collect_index.get(input_handle.solid_name) is None:
-            self._collect_index[input_handle.solid_name] = output_handle
-        else:
-            check.failed("collect index unexpectedly set twice")
+        self._collect_index[input_handle.solid_name].add(output_handle)
 
         # if the output is already fanned out
         if self._dynamic_fan_out_index.get(output_handle.solid_name):
