@@ -1,20 +1,20 @@
 from dagster import check
 from dagster.core.errors import DagsterInvariantViolationError, DagsterRunNotFoundError
-from dagster.core.execution.context.system import SystemExecutionContext
+from dagster.core.execution.context.system import IPlanContext
 from dagster.core.execution.plan.plan import ExecutionPlan
 
 
-def validate_reexecution_memoization(pipeline_context, execution_plan):
-    check.inst_param(pipeline_context, "pipeline_context", SystemExecutionContext)
-    check.inst_param(execution_plan, "execution_plan", ExecutionPlan)
+def validate_reexecution_memoization(
+    plan_context: IPlanContext, execution_plan: ExecutionPlan
+) -> None:
 
-    parent_run_id = pipeline_context.pipeline_run.parent_run_id
+    parent_run_id = plan_context.pipeline_run.parent_run_id
     check.opt_str_param(parent_run_id, "parent_run_id")
 
     if parent_run_id is None:
         return
 
-    if not pipeline_context.instance.has_run(parent_run_id):
+    if not plan_context.instance.has_run(parent_run_id):
         raise DagsterRunNotFoundError(
             "Run id {} set as parent run id was not found in instance".format(parent_run_id),
             invalid_run_id=parent_run_id,
@@ -29,8 +29,6 @@ def validate_reexecution_memoization(pipeline_context, execution_plan):
 
     raise DagsterInvariantViolationError(
         "Cannot perform reexecution with in-memory io managers.\n"
-        "You may have configured non persistent intermediate storage `{}` for reexecution. "
-        "Intermediate Storage is deprecated in 0.10.0 and will be removed in a future release.".format(
-            pipeline_context.intermediate_storage.__class__.__name__
-        )
+        "To enable reexecution, you can set a persistent io manager, such as the "
+        'fs_io_manager, in the resource_defs argument on your ModeDefinition: resource_defs={"io_manager": fs_io_manager}'
     )
